package Sakuin;

use strict;

my $keywords_splitter = '\s*,\s*';
my $phonetic_splitter = "\\|";
my $index_splitter = "\\|";

# Constructor.
sub new {
    my $class = shift;
    my $self = {
                'link_prefix' => '<a href="$link" class="keyword">',
                'link_surfix' => '</a>',
                'link_entry' => 'last'
                };
    
    $self = bless $self, $class;
    $self->cleanup;

    return $self;
}

# Clean up object.
sub cleanup {
    my ($self) = @_;

    $self->{indexes} = {};
    $self->{keywords} = {};
    $self->{phonetics} = {};

    $self->{base_keywords} = {};
    $self->{base_indexes} = [];

    $self->{used_keywords} = [];
    $self->{used_phonetics} = [];
    $self->{used_indexes} = [];

    $self->{related_keywords} = [];
    $self->{related_phonetics} = [];
    $self->{related_indexes} = [];
}

# Set base indexes.
sub set_indexes {
    my ($self, $indexes) = @_;
    my %indexes;
    
    #  Split to lines.
    my @indexes = ();
    my @lines = split("\n", $indexes);
    foreach my $line (@lines) {
        # Split to index and regular expression.
        my @index = split($index_splitter, $line, 2);
        
        # Add index condition.
        $indexes{$index[0]} = $index[1];
        push(@indexes, $index[0]);
    }
    
    # Set as all indexes.
    $self->{indexes} = \%indexes;
    $self->{base_indexes} = \@indexes;
}

# Get index regex.
sub get_index_regex {
    my ($self, $index) = @_;
    return $self->{indexes}->{$index};
}

# Index matching.
sub match_index {
    my ($self, $index, $phonetic) = @_;
    my $regex = $self->get_index_regex($index);
    
    !$regex || $phonetic =~ /$regex/;
}

# Get index for phonetic.
sub get_phonetic_index {
    my ($self, $phonetic) = @_;
    foreach my $index (@{$self->{base_indexes}}) {
        return $index if $self->match_index($index, $phonetic);
    }
}

# Set base keywords.
sub set_base_keywords {
    my ($self, $keywords) = @_;
    my %keywords;
    
    # Split to lines.
    my @lines = split("\n", $keywords);
    foreach my $line (@lines) {
        # Split to keyword and phonetic.
        my @keyword = split($phonetic_splitter, $line, 2);
        
        # Add phonetic.
        $keywords{$keyword[0]} = $keyword[1] if $keyword[1];
    }
    
    # Set as all phonetics.
    $self->{base_keywords} = \%keywords;
}

# Add entry data.
sub add_entry {
    my ($self, $keywords, $id, $title, $link) = @_;
    
    # Build entry structure.
    my $entry = {
            'id' => $id,
            'link' => $link,
            'title' => $title
             };

    #  Split to keywords.
    my @keywords = split($keywords_splitter, $keywords);
    foreach my $keyword (@keywords) {
        # Split to keyword and phonetic.
        my @keyword = split($phonetic_splitter, $keyword, 2);
        my $kw = $keyword[0];
        my $phonetic = $keyword[0];
        
        # Set phonetic.
        if( $keyword[1] ) {
            $phonetic = $keyword[1];
        } elsif ( $self->{base_keywords}->{$kw} ) {
            $phonetic = $self->{base_keywords}->{$kw};
        }
        
        # Get index.
        my $index = $self->get_phonetic_index($phonetic);
        
        # Retrive or create new record.
        my $record;
        if( exists($self->{keywords}->{$kw}) ) {
            $record = $self->{keywords}->{$kw};
        } else {
            $record = {phonetic => $phonetic, entries => [], index => $index};
            $self->{keywords}->{$kw} = $record;
        }

        # Push new entry to entries.
        push(@{$record->{entries}}, $entry);
    }
}

# Get entries of keyword.
sub get_keyword_entries {
    my ($self, $keyword) = @_;
    return $self->{keywords}->{$keyword}->{entries};
}

# Get keyword phonetic.
sub get_keyword_phonetic {
    my ($self, $keyword) = @_;
    return $self->{keywords}->{$keyword}->{phonetic};
}

# Get keyword index.
sub get_keyword_index {
    my ($self, $keyword) = @_;
    return $self->{keywords}->{$keyword}->{index};
}

# Get first entry of keyword.
sub get_first_entry {
    my ($self, $keyword) = @_;
    
    # Get first entry of keyword.
    if( my $entries = $self->get_keyword_entries($keyword) ) {
        return $entries->[0];
    }
}

# Get last entry of keyword.
sub get_last_entry {
    my ($self, $keyword) = @_;
    
    # Get last entry of keyword.
    if( my $entries = $self->get_keyword_entries($keyword) ) {
        my $last = scalar( @$entries ) - 1;
        return $entries->[$last];
    }
}

# Build dictionary after all entries added.
sub build_dictionary {
    my ($self) = @_;
    
    # Sort keywords by length as used keywords.
    $self->{used_keywords} = [sort {length($b) <=> length($a)} keys %{$self->{keywords}}];
    
    # Build used phonetics.
    my %phonetics = ();
    my %indexes = ();
    foreach my $keyword (keys %{$self->{keywords}}) {
        # Retrive or create keyword array of phonetic.
        my $phonetic = $self->get_keyword_phonetic($keyword);
        my $keywords;
        if( exists($phonetics{$phonetic}) ) {
            $keywords = $phonetics{$phonetic};
        } else {
            $keywords = [];
            $phonetics{$phonetic} = $keywords;
        }
        
        # Push to keyword array about phonetic.
        push(@$keywords, $keyword);
        $indexes{$self->get_phonetic_index($phonetic)} = 1;
    }

    $self->{phonetics} = \%phonetics;
    $self->{used_phonetics} = [sort {$a cmp $b} keys %phonetics];

    # Build used indexes.
    my @used_indexes = ();
    foreach my $index (@{$self->{base_indexes}}) {
        if( exists($indexes{$index}) ) {
            # Add to used indexes and break.
            push(@used_indexes, $index);
        }
    }

    $self->{used_indexes} = \@used_indexes;
}

# Replace keyword link.
sub replace_keyword_link {
    my ($self, $rtext, $keyword) = @_;
    
    # Return if text is empty.
    return "" unless $$rtext;
    
    # Return if entry is not found.
    my $entry = (lc($self->{link_entry}) eq 'first')? $self->get_first_entry($keyword): $self->get_last_entry($keyword);
    return "" unless $entry;
    
    # Build anchor tag.
    my %replaces = (
                    '\$title' => $entry->{title},
                    '\$link' => $entry->{link},
                    '\$id' => $entry->{id},
                   );
    
    # Replace prefix.
    my $prefix = $self->{link_prefix};
    foreach my $tag (keys %replaces) {
        $prefix =~ s/$tag/$replaces{$tag}/ig;
    }
    
    # Replace surfix.
    my $surfix = $self->{link_surfix};
    foreach my $tag (keys %replaces) {
        $surfix =~ s/$tag/$replaces{$tag}/ig;
    }
    
    # Replace keyword to link.
    return ($$rtext =~ s/($keyword)/$prefix$1$surfix/g);
}

sub replace_keywords {
    my ($self, $text) = @_;
    my %related_keywords = ();

    # Replace each keywords.
    foreach my $keyword (@{$self->{used_keywords}}) {
        my $res;
        $_ = $text;

        # Check if the keyword is exists at first.
        if( /$keyword/ ) {
            # Find a tag.
            while(/(<)/) {
                $text = substr($_, 0, $-[0]);
                $related_keywords{$keyword} = 1 if $self->replace_keyword_link(\$text, $keyword);
                $res .= $text . $1;
                $_ = substr($_, $+[0]);
                
                if( /^(a [^>]*>)/i ) {
                    # If anchor tag, skip to close tag.
                    $res .= $1;
                    $_ = substr($_, $+[0]);
                    if( /<\/a>/i ) {
                        $res .= substr($_, 0, $+[0]);
                        $_ = substr($_, $+[0]);
                    } else {
                        $res .= $_;
                        $_ = undef;
                        last;
                    }
                } else {
                    # If tag is the other, skip to the end of the tag.
                    if( /^([^>]*>)/i ) {
                        $res .= $1;
                        $_ = substr($_, $+[0]);
                    } else {
                        $res .= $_;
                        $_ = undef;
                        last;
                    }
                }
            }

            # Replace rest of the content.
            $related_keywords{$keyword} = 1 if $self->replace_keyword_link(\$_, $keyword);
            $text = $res . $_;
        }
    }
    
    # Related keywords.
    $self->{related_keywords} = [sort {length($b) <=> length($a)} keys %related_keywords];
    
    # Related phonetics
    my %related_phonetics = ();
    my %related_indexes = ();
    foreach my $keyword (@{$self->{related_keywords}}) {
        # Get phonetic.
        my $phonetic = $self->get_keyword_phonetic($keyword);
        $related_phonetics{$phonetic} = 1;
        $related_indexes{$self->get_phonetic_index($phonetic)} = 1;
    }

    $self->{related_phonetics} = [sort {$a cmp $b} keys %related_phonetics];

    # Build related index.
    my @related_indexes = ();
    foreach my $index (@{$self->{base_indexes}}) {
        if( exists($related_indexes{$index}) ) {
            push(@related_indexes, $index);
        }
    }

    $self->{related_indexes} = \@related_indexes;
    
    $text;
}

# Query indexed keywords from array.
sub query_keywords {
    my ($self, $phonetics, $index) = @_;

    # Loop and push keywords.
    my @keywords = ();
    foreach my $phonetic (@$phonetics) {
        my $keywords = $self->{phonetics}->{$phonetic};
        foreach my $keyword (@{$keywords}) {
            push(@keywords, $keyword) if !$index || $self->get_keyword_index($keyword) eq $index;
        }
    }
    
    \@keywords;
}

# Query indexed and related keywords.
sub query_related_keywords {
    my ($self, $index) = @_;
    $self->query_keywords($self->{related_phonetics}, $index);
}

# Query indexed and used keywords.
sub query_used_keywords {
    my ($self, $index) = @_;
    $self->query_keywords($self->{used_phonetics}, $index);
}

1;